unit HostBattleFormUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Mask, ComCtrls, janTracker, Buttons;

type
  THostBattleForm = class(TForm)
    Label1: TLabel;
    Button1: TButton;
    Button3: TButton;
    PortEdit: TEdit;
    Label2: TLabel;
    PasswordEdit: TEdit;
    Label3: TLabel;
    PlayersTracker: TjanTracker;
    TitleEdit: TEdit;
    Label4: TLabel;
    ModsComboBox: TComboBox;
    Label5: TLabel;
    SpeedButton1: TSpeedButton;
    RankComboBox: TComboBox;
    Label6: TLabel;

    procedure CreateParams(var Params: TCreateParams); override;

    procedure Button3Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure RankComboBoxDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  HostBattleForm: THostBattleForm;

implementation

uses WaitForAckUnit, BattleFormUnit, Unit1, Utility, InitWaitFormUnit,
  DisableUnitsFormUnit, HostInfoUnit, Misc, PreferencesFormUnit;

{$R *.dfm}

procedure THostBattleForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);

  if not Preferences.TaskbarButtons then Exit;

  with Params do begin
    ExStyle := ExStyle or WS_EX_APPWINDOW;
    WndParent := BattleForm.Handle;
  end;
end;

function ValidatePassword(p: string): Boolean;
var
  i: Integer;
begin
  Result := False;

  for i := 1 to Length(p) do if not (p[i] in ['0'..'9', 'a'..'z', 'A'..'Z']) then Exit;

  Result := True;
end;

procedure THostBattleForm.Button3Click(Sender: TObject);
begin
  Close;
end;

procedure THostBattleForm.Button1Click(Sender: TObject);
var
  i: Integer;
  s: string;
begin
  try
    i := StrToInt(PortEdit.Text);
    if (i < 1024) or (i > 65535) then raise Exception.Create('Number out of range');
  except
    MessageDlg('Port number must be a whole number within a proper range (1024..65535)', mtError, [mbOK], 0);
    Exit;
  end;

  if not (PasswordEdit.Text = '') then if not ValidatePassword(PasswordEdit.Text) then
  begin
    MessageDlg('Invalid password! Must contain letters and numbers only!', mtError, [mbOK], 0);
    Exit;
  end;

  if RankComboBox.ItemIndex > 0 then
    if MessageDlg('You should try to avoid limiting games to a certain rank unless it is absolutely necessary'+#13+
                 'as new players will have a limited enough opportunity to play as is.'+#13+#13+
                 'Are you sure you wish to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrNo then Exit;

  if (Status.MyRank < RankComboBox.ItemIndex) then
    if MessageDlg('Your rank (' + Ranks[Status.MyRank] + ') is lower than the battle rank limit (' + Ranks[RankComboBox.ItemIndex] + '). Do you wish to continue anyway?', mtConfirmation, [mbYes, mbNo], 0) = mrNo then Exit;


  // we have to change mod and update unit list:
  InitWaitForm.ChangeCaption(MSG_MODCHANGE);
  InitWaitForm.TakeAction := 0; // change mod
  InitWaitForm.ShowModal; // this changes mod (see OnFormActivate event)
  // now let's update units:
  InitWaitForm.ChangeCaption(MSG_GETUNITS);
  InitWaitForm.TakeAction := 1; // load unit lists
  InitWaitForm.ShowModal; // this loads unit lists (see OnFormActivate event)
  DisableUnitsForm.VDTUnits.RootNodeCount := UnitNames.Count;

  s := 'OPENBATTLE ';
  // password (must be "*" for none):
  if PasswordEdit.Text = '' then s := s + '* ' else s := s + PasswordEdit.Text + ' ';
  // port:
  s := s + PortEdit.Text + ' ';
  // max. players:
  s := s + IntToStr(PlayersTracker.Value) + ' ';
  // metal:
  s := s + IntToStr(BattleForm.MetalTracker.Value) + ' ';
  // energy:
  s := s + IntToStr(BattleForm.EnergyTracker.Value) + ' ';
  // max. units:
  s := s + IntToStr(BattleForm.UnitsTracker.Value) + ' ';
  // start pos.:
  s := s + IntToStr(BattleForm.StartPosRadioGroup.ItemIndex) + ' ';
  // game end condition:
  s := s + IntToStr(BattleForm.GameEndRadioGroup.ItemIndex) + ' ';
  // limit d-gun to startpos:
  s := s + IntToStr(BoolToInt(BattleForm.LimitDGunCheckBox.Checked)) + ' ';
  // hash code:
  s := s + IntToStr(GetModHash(ModsComboBox.Items[ModsComboBox.ItemIndex])) + ' ';
  // rank limit:
  s := s + IntToStr(RankComboBox.ItemIndex) + ' ';
  // map:
  s := s + BattleForm.MapList.Items[BattleForm.MapList.ItemIndex] + #9;
  // title (description):
  if TitleEdit.Text = '' then s := s + '(none)' + #9 else s := s + TitleEdit.Text + #9; // we should not send empty line for title!
  // name of mod:
  s := s + ModsComboBox.Text;

  // send the command:
  MainForm.TryToSendData(s);

  if WaitForAckForm.ShowModal = mrOK then
  begin // success
    Close;
  end
  else
  begin // failure
  end;
end;

procedure THostBattleForm.FormCreate(Sender: TObject);
var
  i: Integer;
begin
  ModsComboBox.Items.Assign(Utility.ModList);
  ModsComboBox.ItemIndex := 0;

  RankComboBox.Items.Add('(No limit)');
  for i := 1 to High(Ranks) do RankComboBox.Items.Add(Ranks[i]);
  RankComboBox.ItemIndex := 0;
end;

procedure THostBattleForm.SpeedButton1Click(Sender: TObject);
begin
  HostInfoForm.ShowModal;
end;

procedure THostBattleForm.RankComboBoxDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  // this ensures the correct highlite color is used
  (Control as TComboBox).Canvas.FillRect(Rect);

  if Index <> 0 then MainForm.RanksImageList.Draw((Control as TComboBox).Canvas, Rect.Left, Rect.Top, Index);
  if Index <> 0 then (Control as TComboBox).Canvas.TextOut(Rect.Left + MainForm.RanksImageList.Width + 3, Rect.Top, (Control as TComboBox).Items[Index])
  else (Control as TComboBox).Canvas.TextOut(Rect.Left + 3, Rect.Top, (Control as TComboBox).Items[Index]);

end;

end.
